@@ -1,13 +1,17 @@
 import os
-from pathlib import PurePosixPath
+from pathlib import Path, PurePosixPath
 from typing import Any, Dict, Iterable, Optional, Union
 
+from botocore.exceptions import ClientError
+
 from ..client import Client, register_client_class
 from ..cloudpath import implementation_registry
 from .s3path import S3Path
 
 try:
     from boto3.session import Session
+    from boto3.s3.transfer import TransferConfig
+    from botocore.config import Config
     import botocore.session
 except ModuleNotFoundError:
     implementation_registry["s3"].dependencies_loaded = False
@@ -24,10 +28,13 @@ class S3Client(Client):
         aws_access_key_id: Optional[str] = None,
         aws_secret_access_key: Optional[str] = None,
         aws_session_token: Optional[str] = None,
+        no_sign_request: Optional[bool] = False,
         botocore_session: Optional["botocore.session.Session"] = None,
         profile_name: Optional[str] = None,
         boto3_session: Optional["Session"] = None,
         local_cache_dir: Optional[Union[str, os.PathLike]] = None,
+        endpoint_url: Optional[str] = None,
+        boto3_transfer_config: Optional["TransferConfig"] = None,
     ):
         """Class constructor. Sets up a boto3 [`Session`](
         https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html).
@@ -43,12 +50,19 @@ class S3Client(Client):
             aws_secret_access_key (Optional[str]): AWS secret access key.
             aws_session_token (Optional[str]): Session key for your AWS account. This is only
                 needed when you are using temporarycredentials.
+            no_sign_request: (Optional[bool]): If `True`, credentials are not looked for and we use unsigned
+                requests to fetch resources. This will only allow access to public resources. This is equivalent
+                to `--no-sign-request` in the AWS CLI (https://docs.aws.amazon.com/cli/latest/reference/).
             botocore_session (Optional[botocore.session.Session]): An already instantiated botocore
                 Session.
             profile_name (Optional[str]): Profile name of a profile in a shared credentials file.
             boto3_session (Optional[Session]): An already instantiated boto3 Session.
             local_cache_dir (Optional[Union[str, os.PathLike]]): Path to directory to use as cache
                 for downloaded files. If None, will use a temporary directory.
+            endpoint_url (Optional[str]): S3 server endpoint URL to use for the constructed boto3 S3 resource and client.
+                Parameterize it to access a customly deployed S3-compatible object store such as MinIO, Ceph or any other.
+            boto3_transfer_config (Optional[dict]): Instantiated TransferConfig for managing s3 transfers.
+                (https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig)
         """
         if boto3_session is not None:
             self.sess = boto3_session
@@ -60,8 +74,23 @@ class S3Client(Client):
                 botocore_session=botocore_session,
                 profile_name=profile_name,
             )
-        self.s3 = self.sess.resource("s3")
-        self.client = self.sess.client("s3")
+
+        if no_sign_request:
+            self.s3 = self.sess.resource(
+                "s3",
+                endpoint_url=endpoint_url,
+                config=Config(signature_version=botocore.session.UNSIGNED),
+            )
+            self.client = self.sess.client(
+                "s3",
+                endpoint_url=endpoint_url,
+                config=Config(signature_version=botocore.session.UNSIGNED),
+            )
+        else:
+            self.s3 = self.sess.resource("s3", endpoint_url=endpoint_url)
+            self.client = self.sess.client("s3", endpoint_url=endpoint_url)
+
+        self.boto3_transfer_config = boto3_transfer_config
 
         super().__init__(local_cache_dir=local_cache_dir)
 
@@ -76,12 +105,11 @@ class S3Client(Client):
             "extra": data["Metadata"],
         }
 
-    def _download_file(
-        self, cloud_path: S3Path, local_path: Union[str, os.PathLike]
-    ) -> Union[str, os.PathLike]:
+    def _download_file(self, cloud_path: S3Path, local_path: Union[str, os.PathLike]) -> Path:
+        local_path = Path(local_path)
         obj = self.s3.Object(cloud_path.bucket, cloud_path.key)
 
-        obj.download_file(str(local_path))
+        obj.download_file(str(local_path), Config=self.boto3_transfer_config)
         return local_path
 
     def _is_file_or_dir(self, cloud_path: S3Path) -> Optional[str]:
@@ -104,17 +132,26 @@ class S3Client(Client):
 
     def _s3_file_query(self, cloud_path: S3Path):
         """Boto3 query used for quick checks of existence and if path is file/dir"""
-        return next(
-            (
-                obj
-                for obj in (
-                    self.s3.Bucket(cloud_path.bucket)
-                    .objects.filter(Prefix=cloud_path.key)
-                    .limit(1)
-                )
-            ),
-            None,
-        )
+        # first check if this is an object that we can access directly
+
+        try:
+            obj = self.s3.Object(cloud_path.bucket, cloud_path.key)
+            obj.load()
+            return obj
+
+        # else, confirm it is a dir by filtering to the first item under the prefix
+        except ClientError:
+            return next(
+                (
+                    obj
+                    for obj in (
+                        self.s3.Bucket(cloud_path.bucket)
+                        .objects.filter(Prefix=cloud_path.key)
+                        .limit(1)
+                    )
+                ),
+                None,
+            )
 
     def _list_dir(self, cloud_path: S3Path, recursive=False) -> Iterable[S3Path]:
         bucket = self.s3.Bucket(cloud_path.bucket)
@@ -151,7 +188,7 @@ class S3Client(Client):
                 for result_key in result.get("Contents", []):
                     yield self.CloudPath(f"s3://{cloud_path.bucket}/{result_key.get('Key')}")
 
-    def _move_file(self, src: S3Path, dst: S3Path) -> S3Path:
+    def _move_file(self, src: S3Path, dst: S3Path, remove_src: bool = True) -> S3Path:
         # just a touch, so "REPLACE" metadata
         if src == dst:
             o = self.s3.Object(src.bucket, src.key)
@@ -165,7 +202,8 @@ class S3Client(Client):
             target = self.s3.Object(dst.bucket, dst.key)
             target.copy({"Bucket": src.bucket, "Key": src.key})
 
-            self._remove(src)
+            if remove_src:
+                self._remove(src)
         return dst
 
     def _remove(self, cloud_path: S3Path) -> None:
@@ -173,12 +211,12 @@ class S3Client(Client):
             obj = self.s3.Object(cloud_path.bucket, cloud_path.key)
 
             # will throw if not a file
-            obj.get()
+            obj.load()
 
             resp = obj.delete()
             assert resp.get("ResponseMetadata").get("HTTPStatusCode") == 204
 
-        except self.client.exceptions.NoSuchKey:
+        except ClientError:
             # try to delete as a direcotry instead
             bucket = self.s3.Bucket(cloud_path.bucket)
 
@@ -196,7 +234,7 @@ class S3Client(Client):
     def _upload_file(self, local_path: Union[str, os.PathLike], cloud_path: S3Path) -> S3Path:
         obj = self.s3.Object(cloud_path.bucket, cloud_path.key)
 
-        obj.upload_file(str(local_path))
+        obj.upload_file(str(local_path), Config=self.boto3_transfer_config)
         return cloud_path
 
 
